#!/usr/bin/env python

import csv
import argparse

import numpy as np
import sys


def generate_discrete_distribution(num_components, rng):
    """
    Generate a discrete probability distribution.
    :param num_components: The number of categories.

    :type numpy.RandomState
    :param rng: A numpy RandomState to generate values from.

    :return: A numpy array with shape (num_components, ) where the components all sum to 1.
    """
    unnormalized_scores = rng.random_sample((num_components,))
    return unnormalized_scores / unnormalized_scores.sum()


def generate_means(num_components, num_features, rng, low=-300, high=300):
    """
    Generate means for Gaussian distributions. A uniform distribution over the range (*low*, *high*) is used to generate
    the components.
    :param num_components: The number of components to generate means for.
    :param num_features: The dimensionality of the means.
    :param rng: A numpy RandomState to generate values from.
    :param low: The start of the random interval.
    :param high: The end of the random interval.
    :return: A numpy array of shape (num_components, num_features).
    """
    return rng.uniform(low, high, (num_components, num_features))

def generate_covs(num_components, num_features, rng, diag=True, low=-50, high=50):
    """
    Generate covariance matrices for Gaussian distributions. The generated matrices will be positive-semidefinite.
    This is done by simply multiplying the matrix by its transpose. The values before the multiplication are drawn from
    a uniform distribution in the range (*low*, *high*).

    :param num_components: The number of covariance matrices to generate.
    :param num_features: The dimension of the covariance matrices
    :param rng: A numpy RandomState to draw paramers from.
    :param diag: If True, the matrices will be diagonal (thus giving independent features), otherwise they will be full.
    :param low: The start of the random interval.
    :param high: The end of the random interval.
    :return: A numpy array of shape (num_components, num_features, num_features)
    """
    random_matrices = rng.uniform(low, high, (num_components, num_features, num_features))
    covs = [np.dot(x, x.T) for x in random_matrices]
    if diag:
        covs = [np.diag(np.diag(cov)) for cov in covs]
    return np.array(covs)


def gmm(component_probs, means, covs, num_samples, rng=None, include_labels=True):
    """
    Generate samples from a gaussian mixture model with the given parameters.
    :param component_probs: A 1D array of component probabilities.
    :param means: A 2D array of means for each component of shape (num_components, num_features)
    :param covs: A 3D array of covariance matrices for the components of shape (num_components, num_features, num_features)
    :param num_samples: The number of samples to generate.
    :param rng: Either a numpy.random.RandomSeed object or a constant to use for seeding such an object. Will be used
                as the random number generator for the samples.
    :param include_labels: If True, the class labels for the samples is also returned. If false, only the
                           samples are returned.
    :return: If include_labels is False, returns a 2D numpy array of samples, where each row is a different sample and
             the columns are the features. If True, returns a pair of numpy arrays (class_labels, samples), where
             class_labels is a numpy array with the same length as samples, where each element is the component label
             for the corresponding row in *samples*.
    """
    if not isinstance(rng, np.random.RandomState):
        rng = np.random.RandomState(rng)

    num_components, num_features = means.shape
    samples = np.zeros((num_samples, num_features))

    classes = rng.choice(num_components, size=num_samples, p=component_probs)

    for i in range(num_components):
        (component_indices,) = (classes == i).nonzero()
        component_samples = rng.multivariate_normal(mean=means[i], cov=covs[i], size=len(component_indices))
        samples[component_indices] = component_samples

    if include_labels:
        return (classes, samples)
    else:
        return samples


def main():
    parser = argparse.ArgumentParser(description="Script for generating data from a mixture of gaussians")
    parser.add_argument('num_components',
                        help=("The number of components to use. All other arguments will be parsed with this number "
                              "in mind."),
                        type=int)
    parser.add_argument('num_features',
                        help=("The number of features to use. All other arguments will be parsed with this number "
                              "in mind."),
                        type=int)
    parser.add_argument('num_samples',
                        help=("The number of samples to draw."),
                        type=int)
    parser.add_argument('--unlabeled',
                        help=("If this flag is set, the samples will not have a class label."),
                        action='store_true')
    parser.add_argument('--output', '-o',
                        help=("Output file. If omitted samples are dumped to standard out."),
                        type=argparse.FileType('w'),
                        default=sys.stdout)

    parser.add_argument('--parameter-seed',
                        help=("Constant to seed the random number generator with when creating the mixture. "
                              "This way the same mixture model is generated. By default, the seed is constant"
                              "so that randomly generated parameters for the mixture stays the same. This doesn not"
                              "affect the random number generator for the samples."),
                        type=int,
                        default=1729)

    parser.add_argument('--component-probs',
                        help=("The probabilities for each of the classes. This will be normalized to sum up to 1, so"
                              "the values don't have to constitute a proper probability distribution. num_components"
                              " arguments must be supplied. If this argument is omitted, the class probabilities will"
                              " be randomly generated."),
                        nargs='+',
                        type=float)
    parser.add_argument('--means',
                        help=("List of means for the distributions. These will be partitioned into num_components "
                              "seperate vectors of num_features length. You must supply num_components * num_feature "
                              "arguments. If this argument is not supplied, random numbers will be generated for the "
                              "means"),
                        nargs='+',
                        type=float)
    parser.add_argument('--covs',
                        help=("List of covariance matrix entries for the distributions. These will be partitioned into "
                              "num_components seperate vectors. If the partitions have num_feature lengths, they are"
                              "assumed to be the diagonal elements of the covariance matrix. If the are num_feature * "
                              "num_features elements of each partitions, they will be used as a full covariance matrix."
                              " If this argument is omitted, random numbers will be generated for the covariance "
                              "matrices. The matrices must be symmetric and positive-semidefinite"),
                        nargs='+',
                        type=float)
    parser.add_argument('--diag', help="If the covariance matrices are randomly generated, this toggles whether they"
                                       "are diagonal or full.", action='store_true')

    args = parser.parse_args()
    num_components = args.num_components
    num_features = args.num_features

    parameter_rng = np.random.RandomState(args.parameter_seed)

    if args.component_probs is None:
        component_probs = generate_discrete_distribution(num_components, parameter_rng)
    else:
        if len(args.component_probs) != num_components:
            raise ValueError("The number of arguments to --component-probs must equal num_components.")
        component_probs = args.component_probs

    if args.means is None:
        means = generate_means(num_components, num_features, parameter_rng)
    else:
        if len(args.means) != num_features * num_components:
            raise ValueError("The number of arguments to --means must equal num_components * num_features.")
        means = np.array(args.means).reshape(num_components, num_features)

    if args.covs is None:
        covs = generate_covs(num_components, num_features, parameter_rng, diag=args.diag)
    else:
        if len(args.covs) != num_components * num_features * num_features:
            # Full matrices
            covs = np.array(args.covs).reshape(num_components, num_features, num_features)
        elif len(args.covs) != num_components * num_features:
            # Diagonal matrices
            variances = np.array(args.means).reshape(num_components, num_features)
            # variances is now an ndarray where the first axis indexes the components. We multiply each of the columns
            # with an identity matrix and get diagonal matrices
            covs = [np.eye(num_features) * v for v in variances]

        else:
            raise ValueError("The number of arguments to --covs must equal num_components * num_features or "
            "num_components * num_features * num_features.")

    print("Component probabilities: {}".format(component_probs))
    print("Means: {}".format(means))
    print("Covariance matrices: {}".format(covs))

    columns = [] #[list(range(args.num_samples))]
    fieldnames = [] #['id']
    if args.unlabeled:
        samples = gmm(component_probs, means, covs, args.num_samples, include_labels=False)
        # Splits the columns into independent arrays.
        columns += [sample.flatten() for sample in np.hsplit(samples, num_features)]
        fieldnames += ['x{}'.format(i) for i in range(num_features)]
    else:
        class_labels, samples = gmm(component_probs, means, covs, args.num_samples, include_labels=True)
        # Splits the columns into independent arrays.
        columns += [class_labels] + [sample.flatten() for sample in np.hsplit(samples, num_features)]
        fieldnames += ['component'] + ['x{}'.format(i) for i in range(num_features)]

    csv_writer = csv.writer(args.output)
    csv_writer.writerow(fieldnames)
    csv_writer.writerows(zip(*columns))  # This essentially transposes the list of columns to a iterator over rows


if __name__ == '__main__':
    main()