package se.sics.streamingclassifier

import org.apache.flink.streaming.api.functions.co.CoFlatMapFunction
import org.apache.flink.streaming.api.windowing.StreamWindow
import org.apache.flink.util.Collector

/**
 * The flat map operator which is applied to the combined training and test stream. Training examples are consumed as
 * soon as they arrive and updates the model while test data is consumed in a sliding window.
 * @param model: The classification model to train and use for prediction
 */
class ClassifierOperator(val model: ClassificationModel) extends CoFlatMapFunction[LabeledSample, StreamWindow[LabeledSample], PredictionResults] {
  var totalStream1Seen = 0
  var totalStream2Seen = 0

  /**
   * flatMap operator for the training stream.
   * @param labeledSample: A labeled sample from the training stream.
   * @param out: Collector for producing the stream output. This flat map operator never puts anything in the collector.
   */
  override def flatMap1(labeledSample: LabeledSample, out: Collector[PredictionResults]): Unit = {
    totalStream1Seen += 1
    model.update(labeledSample)
  }

  /**
   * flatMap operator for the test stream. The current model is used to produce a prediction for each of the
   * samples in a window. The results of the predictions is summarized in a confusion matrix together with some
   * more informative statistics.
   * @param labeledWindow: A StreamWindow of labeled example on which the model will be used to create predictions.
   * @param out: A Collector which is used for the stream output. Each window generates a single PredictionResults
   *             object which summarizes the performance of the model.
   */
  override def flatMap2(labeledWindow: StreamWindow[LabeledSample], out: Collector[PredictionResults]): Unit = {
    totalStream2Seen += 1

    //Since we might have more classes in the model than there actually are classes, we check which ones are
    // actually used in the data. actualClasses keeps track of this.
    var actualClasses: Set[ClassLabel] = Set()
    val predictions: Array[LabeledPrediction] = labeledWindow.toArray.map {
      case LabeledSample(label, features) =>
        val predictedLabel = model.predict(UnlabeledSample(features))
        actualClasses += label
        actualClasses += predictedLabel
        LabeledPrediction(label, predictedLabel)
    }
    val numClasses = actualClasses.max + 1  // +1 since the class labels are 0-indexed

    // We calculate the statistics by accumulating the predictions in a PredictionResults object
    val initialResults = new PredictionResults(numClasses)
    def folder(previousResults: PredictionResults, prediction: LabeledPrediction): PredictionResults =
      previousResults.addPrediction(prediction)
    val predictionResults: PredictionResults = predictions.foldLeft(initialResults)(folder)

    // We add some extra data to the object which will be printed to the stream.
    predictionResults.testWindowId = totalStream2Seen
    predictionResults.totalTrainingCount = totalStream1Seen

    //This is some helpful debug output
    predictionResults.comments = {
      val nbModel = model.asInstanceOf[NaiveBayesModel]
      modelParameterSummary(nbModel, numClasses)
    }

    out.collect(predictionResults)
  }

  /**
   * Returns a string with model parameters for a NaiveBayesModel.
   * @param model: The model object to get parameters from.
   * @param numClasses: The number of classes to include in the output. The model might have been initialized with a
   *                    higher number of classes than are actually present.
   * @return A string with the essential model parameters given.
   */
  def modelParameterSummary(model: NaiveBayesModel, numClasses: Int): String = {
    /**
     * Summarizes the model parameters in a string. Useful for printf debugging.
     */
    val indent = "\n    "

    val probString = model.classStatistics.take(numClasses).zipWithIndex.map {
      case (stats, i) =>
        val classSummary = f"$i: (${stats.classProbability})"
        classSummary
    }.mkString(indent)

    val meansString = model.classStatistics.take(numClasses).zipWithIndex.map {
      case (stats, i) =>
        val featureString = stats.mean.map(x => f"$x%.3f").mkString(", ")
        val classSummary = f"$i: ($featureString)"
        classSummary
    }.mkString(indent)

    val varsString = model.classStatistics.take(numClasses).zipWithIndex.map {
      case (stats, i) =>
        val featureString = stats.variance.map(x => f"$x%.3f").mkString(", ")
        val classSummary = f"$i: ($featureString)"
        classSummary
    }.mkString(indent)

    { f"Class probabilities:" + indent + f"$probString%s\n" +
      f"Class means:" + indent + f"$meansString%s\n" +
      f"Class variances:" + indent + f"$varsString%s"}
  }

}